#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/poll.h>

extern "C" {
#include "md5.h"
}

#include "common.h"

#define DEBUG

/*************************** PROTOCOL MESSAGES **************************/

static char *Messages [BADMSG];

int initLOCMessages ()
{
	// This function is CRITICAL.
	// All programs must AGREE on the message strings!
	Messages [VERSION] 		= "LOC";
	Messages [IAM] 			= "IAM";
	Messages [AUTHSTR]		= "AUT";
	Messages [DIGEST]		= "DIG";
	Messages [OKEY]			= "OK ";
	Messages [DISCONNECT]		= "HUP";
	Messages [USERADDRESS]		= "UAD";
	Messages [FIREWALLED]		= "FRW";
	Messages [CHPASS]		= "CPA";
	Messages [WANNATALKTO]		= "WTT";
	Messages [GOTOADDRESS]		= "GTA";
	Messages [SETUPRENDEZVOUS]	= "RDZ";
	Messages [NOTCONNECTED]		= "NOC";
	Messages [ADMIN]		= "ADM";
	Messages [AD_ADDUSER]		= "ADD";
	Messages [AD_DELUSER]		= "DEL";
	Messages [AD_CHPASS]		= "ACP";
	Messages [USEREXISTS]		= "UEX";
	Messages [NOSUCHUSER]		= "NOU";
	Messages [BADPARAMETER]		= "BPA";
	return (int) BADMSG;
}

LOCMSG MSGtype (char *s)
{
	int i;

	for (i = (int)VERSION; i < BADMSG; i++)
		if (!strcmp (s, Messages [i])) return (LOCMSG)i;

	return (LOCMSG)i;
}

// Generate a LOC message in <ret> which must be long enough.
// arguments are separated with \0, so the resulting string is
// not null-terminated.
// Return value: the size of the resulting message

int mkLOCmessage (char *ret, LOCMSG l, int argc, char **argv)
{
	char *r = ret;

	r += strlen (strcpy (r, Messages [l])) + 1;

#ifdef DEBUG
	fprintf (stderr, "+++++++++++++++++Creating Message:\n");
	fprintf (stderr, "Message [%s] ", Messages [l]);
	for (int i = 0; i < argc; i++)
		fprintf (stderr, " [%s] ", argv [i]);
	fprintf (stderr, "\n");
#endif
	for (int i = 0; i < argc; i++)
		r += strlen (strcpy (r, argv [i])) + 1;

#ifdef DEBUGS
	fprintf (stderr, "Final message size [%i]\n", r - ret);
#endif

	return r - ret;
}

// parse a LOC message's arguments in <argv> which must be
// long enough (3 fields max at the moment).
// Return value: the number of arguments

int parseLOCmessage (char *rcvd, int l, char **argv)
{
	int p, i;

	p = i = 0;
	while (i < MAX_ARGV && p < l) {
		argv [i] = rcvd + p;
		p += strlen (argv [i]) + 1;
		++i;
	}

#ifdef DEBUG
	fprintf (stderr, "-------Pck [%i] cntns %i args:", l, i);
	for (p = 0; p < i; p++)
		fprintf (stderr, "[%s] ", argv [p]);
	fprintf (stderr, "\n");
#endif

	return i;
}

/************************* COMMON DIGEST GENERATION ***********************/
// Produce md5 digest in hex string form. <ret> must be char [33]
// Thanks to Ulrich Drepper @ GNU for the md5.[ch]

char *do_md5digest (char *s1, char *s2, char *ret)
{
	md5_ctx context;
	unsigned char out [16];
	char tmp [3];
	int i;
	char *data = (char*) alloca (strlen (s1) + strlen (s2) + 1);

	strcat (strcpy (data, s1), s2);
	md5_init_ctx (&context);
	md5_process_bytes (data, strlen (data), &context);
	md5_finish_ctx (&context, out);

	*ret = 0;
	for (i = 0; i < 16; i++) {
		sprintf (tmp, "%02x", out [i]);
		strcat (ret, tmp);
	}

	return ret;
}
