
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/types.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/poll.h>
#include <string.h>
#include <ctype.h>
#include <netdb.h>
#include <sys/time.h>
#include <unistd.h>
#include <pthread.h>
#include <semaphore.h>

#include "common.h"
#include "dbstree.h"

#define DEBUG

// The infamous

char *StrDup (char *c)
{
	return strcpy (new char [strlen (c) + 1], c);
}

/****************************** UTILITIES MISC ***************************/

#define RSTRING_SIZE_UINT 3
typedef char randstring [RSTRING_SIZE_UINT * 6 + 3];

// Base64 encoding. <Encoded> is where the encoded string will be written.
// this is 3 bytes of data produce 4 bytes of Base64 text.

static const char Base64Table [] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=";

static char *Encode64 (char *p, char *Encoded)
{
	char *e;

	e = Encoded;

	while (1)
	{
		if (*p == 0) break;
		*e++ = (*p & 252) >> 2;
		*e++ = ((*p & 3) << 4) + ((*++p & 240) >> 4);
		if (*p == 0) 
		{
			*e++ = 64;
			*e++ = 64;
			break;
		}
		*e++ = ((*p & 15) << 2) + ((*++p & 192) >> 6);
		if (*p == 0) {
			*e++ = 64;
			break;
		} else *e++ = (*p & 63);
		p++;
	}
	
	*e = 65;
	
	for (p = Encoded; p <= e ; p++)
		*p = Base64Table [(int) *p];

	return Encoded;
}

// We don't want predictable random numbers.
// cryptorand() returns hashed random values so theoretically if you
// know a cryptorand you cannot find the next cryptorand.
// Brute force attack _is_ possible but its not worth it.
// The hashing is [X**rand()], where X a random odd number
// less than 100 and not 1. This operation is truncated at 32-bits
// so it is modulo 2^32.

static unsigned int xyc (unsigned int x, unsigned int y)
{
	x += 2;
	x |= 1;
	unsigned int s, t, u;

	s = 1; t = x; u = y;

	while (u) {
		if (u & 1) s = s * t;
		u>>=1;
		t = t*t;
	}

	return s;
}

static unsigned int cryptorand ()
{
	return xyc (rand ()%100, rand()) ^ rand ();
}

// generate a random string RSTRING_SIZE_UINT integers long.
// the procedure is to generate cryptorands and convert to
// Base 64 string.

static char *rand_string (randstring ret)
{
	unsigned int array [RSTRING_SIZE_UINT + 1];
	int i;

	for (i = 0; i < RSTRING_SIZE_UINT; i++)
		array [i] = cryptorand ();
	array [i] = 0;

	Encode64 ((char*) array, ret);

	return ret;
}

/****************************** COMMON POINT *****************************/

// Set up a rendezvous point for a firewalled client.
// The two clients are notified about the <inet-point> and
// a <sesame> word.
// Thread remains as a proxy as long as two attendants need it,
// so arrange() must be called by a thread

static int listen_tcp (short int *port)
{
	int fd;
	struct sockaddr_in socka;
	struct sockaddr_in porter;
	socklen_t j;

	fd = socket (AF_INET, SOCK_STREAM, 0);

	socka.sin_family = AF_INET;
	socka.sin_port = 0;
	socka.sin_addr.s_addr = htonl (INADDR_ANY);

	if (bind (fd, (struct sockaddr*)&socka, sizeof socka))
		return -1;
	listen (fd, 2);

	j = sizeof (porter);
	if (getsockname (fd, (struct sockaddr*)&porter, &j) < 0)
		return -1;

	*port = ntohs (porter.sin_port);

	return fd;
}

#define TIMEOUT 300*1000
#define BUFFERSIZE 1024

static int getpass (struct pollfd *pf, char *pass)
{
	char *c = (char*) alloca (strlen (pass) + 1);
	int i = poll (pf, 1, TIMEOUT);
	if (i == 0) return -1;
	i = read (pf->fd, c,  strlen (pass));
	c [i] = 0;
	pf->revents = 0;
	return strcmp (pass, c);
}

struct fdpass
{
	int fd;
	char pass [10];
};

static void *compoint (void *v)
{
#define PV ((fdpass*)v)
	int i = 0;
	socklen_t S;
	struct pollfd fds [2];
	struct sockaddr_in foo [2];
	char Buffer [BUFFERSIZE];

	if ((fds [0].fd = accept (PV->fd, (struct sockaddr*)&foo[0], &S)) == -1)
	{
		close (PV->fd);
		return NULL;
	}

	if ((fds [1].fd = accept (PV->fd, (struct sockaddr*)&foo[1], &S)) == -1)
	{
		close (PV->fd);
		close (fds [0].fd);
		free (v);
		return NULL;
	}

	close (PV->fd);

	fds [0].events = fds [1].events = POLLIN;
	fds [0].revents = fds [1].revents = 0;

	if (getpass (&fds [0], PV->pass) || getpass (&fds [1], PV->pass)) {
		close (fds [0].fd);
		close (fds [1].fd);
		free (v);
		return NULL;
	}

	while (1) {
		i = poll (fds, 2, TIMEOUT);
		if (i == -1) {
			perror ("poll");
			return NULL;
		}
		if (i == 0) break;
		if (fds [0].revents) {
			i = read (fds [0].fd, Buffer, BUFFERSIZE);
#ifdef DEBUG
Buffer [i]=0;
fprintf (stderr, "Received %i [%s]\n", i, Buffer);
#endif
			if (i > 0)
				write (fds [1].fd, Buffer, i);
			else break;
			fds [0].revents = 0;
		}
		if (fds [1].revents) {
			i = read (fds [1].fd, Buffer, BUFFERSIZE);
#ifdef DEBUG
Buffer [i]=0;
fprintf (stderr, "Received %i [%s]\n", i, Buffer);
#endif
			if (i > 0)
				write (fds [0].fd, Buffer, i);
			else break;
			fds [1].revents = 0;
		}
	}

#ifdef DEBUG
fprintf (stderr, "# # # # # # # # # # COMPOINT Closed # # # # # # #\n");
#endif
	close (fds [0].fd);
	close (fds [1].fd);

	delete (fdpass*)v;

	return NULL;
}

static void arrange (int cli1, int cli2)
{
	short int P;
	fdpass *fp = new fdpass;

	sprintf (fp->pass, "%x", cryptorand ());
	fp->fd = listen_tcp (&P);
	
	char N [10], MSG [MAX_MSG+2], *argv [2] = { N, fp->pass };
	sprintf (N, "%i", P);

	P = mkLOCmessage (MSG, SETUPRENDEZVOUS, 2, argv);
	
	write (cli1, MSG, P);
	write (cli2, MSG, P);

	// *** If we want to fork off proxy thread
	//pthread_t TH;
	//pthread_create (&TH, NULL, compoint, fp);
	//pthread_detach (TH);
	// *** else
	compoint (fp);
}

/****************************** USER DATABASE ****************************/

// thread safe user database entries
#define RACTION_ADD 'A'
#define RACTION_DEL 'D'
#define RACTION_CHP 'C'

static dbsTree UserTree;

struct CON {
	int type;
};

class User : public dbsNode
{
   public:
	char *Password;
	CON *Conndata;
	User (char *u, char *p) : dbsNode (&UserTree)
		{ Name = StrDup (u); Password = StrDup (p); Conndata = NULL; }
	~User ()	{ UserTree.dbsRemove (this); }
};

static sem_t asem, rsem;
static int recovery_fd;

static void recovery_data (char a, char *s1, char *s2 = NULL)
{
	char buffer [200];	// supeflouus
	int i;

	if (s2) i = sprintf (buffer, "%c\n%s\n%s\n", a, s1, s2);
	else i = sprintf (buffer, "%c\n%s\n", a, s1);

	sem_wait (&rsem);
	write (recovery_fd, buffer, i);
	sem_post (&rsem);
}

/*
 Only one semaphore. normally there should be read/write access and many
 semaphores for multithreaded access of the user database.
 However, all actions are log(N)-non-blocking [Except 'Save' DB].
 log(N) in modern processors is instant.
*/

static void load_database (FILE *fd)
{
	sem_init (&asem, 0, 1);
	sem_init (&rsem, 0, 1);

	if (fd) {
		char Name [MAX_NAME], Pass [MAX_PASS], *r;
		for (;;) {
			r = fgets (Name, MAX_NAME, fd);
			if (!r) break;
			r = fgets (Pass, MAX_PASS, fd);
			if (!r) break;
			Name [strlen (Name) - 1] = 0;
			Pass [strlen (Pass) - 1] = 0;
			if (UserTree.dbsFind (Name)) continue;
			new User (Name, Pass);
		}
		printf ("%i user accounts\n", UserTree.nnodes);
	}
}

static char *passof (char *u, char *ret)
{
	sem_wait (&asem);
	User *U;

	if ((U = (User*) UserTree.dbsFind (u))) u = strcpy (ret, U->Password);
	else u = NULL;

	sem_post (&asem);

	return u;
}

static bool chpass (char *u, char *p)
{
	sem_wait (&asem);
	User *U;
	bool r;

	if (r = (U = (User*) UserTree.dbsFind (u)) != NULL) {
		delete [] U->Password;
		U->Password = StrDup (p);
	}

	sem_post (&asem);

	if (r) recovery_data (RACTION_CHP, u, p);

	return r;
}

static bool adduser (char *u, char *p)
{
	sem_wait (&asem);

	if (UserTree.dbsFind (u)) {
		sem_post (&asem);
		return false;
	}

	new User (u, p);

	sem_post (&asem);

	recovery_data (RACTION_ADD, u, p);

	return true;
}

static FILE *savedb;
static void savenode (dbsNode *n)
{
	fprintf (savedb, "%s\n%s\n", ((User*)n)->Name, ((User*)n)->Password);
}

static void savedata (FILE *fd)
{
	// the caller of this must bring down the semaphores
	savedb = fd;
	UserTree.copytree (savenode);
}

/*****************/

#define TYPE_TCPIP 1
#define TYPE_FIREWALLED 2

struct CON_TCPIP : CON {
	unsigned int IP;
	short port;
	CON_TCPIP (unsigned int I, short P)
	{ type = TYPE_TCPIP; IP = I; port = P; }
};

struct CON_FIREWALLED : CON {
	User *U;
	int fd;
	CON_FIREWALLED (User *u, int f)
	{ type = TYPE_FIREWALLED; U = u; fd = f; }
	void forkloop ();
};

/*****************/

static void disconnect (CON *C)
{
	switch (C->type) {
	case TYPE_TCPIP:
		delete (CON_TCPIP*) C;
		break;
	case TYPE_FIREWALLED:
		// close the file descriptor. poll() should then
		// terminate. The poll-thread should then delete
		// itself (CON_FIREWALLED) and set the User state
		// to disconnected. Here, by setting U to NULL
		// we avoid that.
		((CON_FIREWALLED*)C)->U = NULL;
		close (((CON_FIREWALLED*)C)->fd);
		break;
	}
}

static bool deluser (char *u)
{
	sem_wait (&asem);
	User *U;

	if (!(U = (User*) UserTree.dbsFind (u))) {
		sem_post (&asem);
		return false;
	}

	if (U->Conndata)
		disconnect (U->Conndata);
	
	delete U;

	sem_post (&asem);

	recovery_data (RACTION_DEL, u);

	return true;
}

static bool userconnect (char *u, unsigned int IP, short port)
{
	sem_wait (&asem);
	User *U;

	if (!(U = (User*) UserTree.dbsFind (u))) {
		sem_post (&asem);
		return false;
	}

	if (U->Conndata)
		disconnect (U->Conndata);
	
	U->Conndata = new CON_TCPIP (IP, port);

	sem_post (&asem);

	return true;
}

static void userconnect (char *u, int fd)
{
	sem_wait (&asem);
	User *U;

	if (!(U = (User*) UserTree.dbsFind (u))) {
		sem_post (&asem);
		return;
	}

	if (U->Conndata)
		disconnect (U->Conndata);
	
	CON_FIREWALLED *CF = new CON_FIREWALLED (U, fd);
	U->Conndata = (CON*) CF;

	sem_post (&asem);

	CF->forkloop ();

	if (CF->U) {
		sem_wait (&asem);
		// set to disconnected, unless we where explictly disconnected
		// for a new connection request.
		CF->U->Conndata = NULL;
		sem_post (&asem);
	}
}

static void userhup (char *u)
{
	sem_wait (&asem);
	User *U;

	if ((U = (User*) UserTree.dbsFind (u)) && U->Conndata)
		disconnect (U->Conndata);

	sem_post (&asem);
}

static bool wannatalk (int fd, char *u)
{
	sem_wait (&asem);
	User *U;

	if (!(U = (User*) UserTree.dbsFind (u)) || !U->Conndata) {
		sem_post (&asem);
		return false;
	}

	switch (U->Conndata->type) {
	case TYPE_TCPIP: {
		short port = ((CON_TCPIP*)U->Conndata)->port;
		struct in_addr IP;
		int l;
		char ps [10], *argv [2], msgbuf [MAX_MSG];

		IP.s_addr = ((CON_TCPIP*)U->Conndata)->IP;

		sem_post (&asem);
		sprintf (ps, "%u", port);
		argv [0] = inet_ntoa (IP);
		argv [1] = ps;
		l = mkLOCmessage (msgbuf, GOTOADDRESS, 2, argv);

		write (fd, msgbuf, l);
		} break;
	case TYPE_FIREWALLED: {
		int fd2 = ((CON_FIREWALLED*)U->Conndata)->fd;

		sem_post (&asem);

		arrange (fd, fd2);
		} break;
	default: sem_post (&asem);
	}

	return true;
}

static void sendAck (int, LOCMSG);

void CON_FIREWALLED::forkloop ()
{
	int i;
	struct pollfd pfd;
	char Buff [10];

	sendAck (fd, OKEY);

	pfd.fd = fd;
	pfd.events = POLLIN;
	pfd.revents = 0;

	for (;;) {
		i = poll (&pfd, 1, -1);
		if (i == 0) break;
		if (pfd.revents & ~POLLIN)
			break;			// pollhup or pollerr
		pfd.revents = 0;
		i = read (pfd.fd, Buff, 10);
		if (i == 0) break;		// pollin zero==HUP
	}
}

static bool valid_string (char *s, int max)
{
	int i, j = strlen (s);
	if (j >= max) return false;

	for (i = 0; i < j; i++)
		if (s [i] < 32) break;

	return i == j;
}

/********************** SERVER FUNCTIONS ***************************/

static char VERSION_STR [10];
static int VERSION_LEN;
static char *admin_password;

static void server_init ()
{
	char *argv [1] = { LOCSERVER_VERSION };
	VERSION_LEN = mkLOCmessage (VERSION_STR, VERSION, 1, argv);

}

static int listen_tcp (short int port)
{
	int fd;
	struct sockaddr_in socka;

	fd = socket (AF_INET, SOCK_STREAM, 0);

	socka.sin_family = AF_INET;
	socka.sin_port = htons (port);
	socka.sin_addr.s_addr = htonl (INADDR_ANY);

	if (bind (fd, (struct sockaddr*)&socka, sizeof socka))
		return -1;
	listen (fd, 30);

	return fd;
}

#define CLIENT_TIMEOUT 40*1000

static int getmessage (char *ret, int fd, unsigned int timeout = CLIENT_TIMEOUT)
{
	struct pollfd pfd  = { fd, POLLIN, 0 };
	int i;

	i = poll (&pfd, 1, CLIENT_TIMEOUT);
	if (i != 1) {
		close (fd);
		return -1;
	}

	i = read (fd, ret, MAX_MSG-2);

	if (i == 0 || i > MAX_MSG-2) {
		// yes we are rude. very rude. If the message
		// is longer than the well-estimated MAX_MSG
		// we shut the client down. Just like that.
		// Thank you.
		close (fd);
		return -1;
	}

	return i;
}

static bool MD5digestOK (char *s1, char *s2, char *d)
{
	md5string M;
	return strcmp (do_md5digest (s1, s2, M), d) == 0;
}

static void sendAck (int fd, LOCMSG m)
{
	char buffer [MAX_MSG];
	int i;
	i = mkLOCmessage (buffer, m, 0, NULL);
	write (fd, buffer, i);
}

static bool CHAP (char *pass, int fd)
{
	randstring R;
	int i;
	char *argv [MAX_ARGV + 1], buffer [MAX_MSG];

	argv [0] = rand_string (R);
	i = mkLOCmessage (buffer, AUTHSTR, 1, argv);
	write (fd, buffer, i);
	if ((i = getmessage (buffer, fd)) == -1) return false;
	i = parseLOCmessage (buffer, i, argv);
	if (i != 2 || MSGtype (argv [0]) != DIGEST) return false;

	return MD5digestOK (pass, R, argv [1]);
}

#define MSGcase(X, Y) case X: if (i != Y) break;

static void authenticate (int fd, char *u)
{
	int i;
	char *argv [MAX_ARGV+1], buffer [MAX_MSG+1], pass [MAX_PASS+1];

	if (!passof (u, pass)) return;
	if (!CHAP (pass, fd)) return;

	sendAck (fd, OKEY);

	if ((i = getmessage (buffer, fd)) == -1) return;

	i = parseLOCmessage (buffer, i, argv);

	if (i > 0) switch (MSGtype (argv [0])) {
	MSGcase(DISCONNECT, 1)
		userhup (u);
		break;
	MSGcase(USERADDRESS, 3) {
		struct in_addr IP;
		if (inet_aton (argv [1], &IP))
			userconnect (u, IP.s_addr, atoi (argv [2]));
			sendAck (fd, OKEY);
		} break;
	MSGcase(FIREWALLED, 1)
		userconnect (u, fd);
		break;
	MSGcase(CHPASS, 2)
		if (valid_string (argv [1], MAX_PASS))
			if (chpass (u, argv [1]))
				sendAck (fd, OKEY);
			else sendAck (fd, NOSUCHUSER);
		else sendAck (fd, BADPARAMETER);
	default:
		break;
	}
}

static void administrator (int fd)
{
	int i;
	char *argv [MAX_ARGV+1], buffer [MAX_MSG];

	if (!CHAP (admin_password, fd)) return;
	sendAck (fd, OKEY);

	while (1) {
		if ((i = getmessage (buffer, fd, 0)) == -1)
			break;
		i = parseLOCmessage (buffer, i, argv);
		if (i > 0) switch (MSGtype (argv [0])) {
		MSGcase(AD_ADDUSER, 3)
			if (valid_string (argv [1], MAX_NAME)
			&& valid_string (argv [2], MAX_PASS))
				if (adduser (argv [1], argv [2]))
					sendAck (fd, OKEY);
				else sendAck (fd, USEREXISTS);
			else sendAck (fd, BADPARAMETER);
			break;
		MSGcase(AD_DELUSER, 2)
			if (deluser (argv [1]))
				sendAck (fd, OKEY);
			else sendAck (fd, NOSUCHUSER);
			break;
		MSGcase(AD_CHPASS, 3)
			if (valid_string (argv [1], MAX_NAME)
			&& valid_string (argv [2], MAX_PASS))
				if (chpass (argv [1], argv [2]))
					sendAck (fd, OKEY);
				else sendAck (fd, NOSUCHUSER);
			else sendAck (fd, BADPARAMETER);
		default: // admin gone mad???
			break;
		}
	}
}

static void *do_client (void *v)
{
	int *f = (int*) v;
	int FD = *f;
	delete f;

	int i;
	char *argv [MAX_ARGV], buffer [MAX_MSG];

	write (FD, VERSION_STR, VERSION_LEN);

	if ((i = getmessage (buffer, FD)) == -1) return NULL;

	i = parseLOCmessage (buffer, i, argv);

	if (i > 0) switch (MSGtype (argv [0])) {
	MSGcase(WANNATALKTO, 2)
		if (!wannatalk (FD, argv [1])) {
			i = mkLOCmessage (buffer, NOTCONNECTED, 0, NULL);
			write (FD, buffer, i);
		}
		break;
	MSGcase(IAM, 2)
		authenticate(FD, argv[1]);
		break;
	MSGcase(ADMIN, 1)
		administrator(FD);
	default:
		break;
	}

	close (FD);
	return NULL;
}

static int server (int sockfd)
{
	int ret, clifd;
	sockaddr_in client;
	pthread_t pt;
	int *fd;
	socklen_t stoopid;

	while (1) {
		clifd = accept (sockfd, (sockaddr*)&client, &stoopid);
		if (clifd == -1) continue;

		fd = new int (clifd);

		ret = pthread_create (&pt, NULL, do_client, fd);

		if (ret == 0) pthread_detach (pt);
		else fprintf (stderr, "Can't create client thread\n");
	}

	return 0;
}

/********************************* MAIN **********************************/

#define DEFAULT_DIR "/.LOCserv"
#define DEFAULT_PORT 9090
#define LOCKFILE "lock"
#define ADMINPASSFILE "admin_passwd"
#define DBFILE "users"
#define RECOVERYFILE "recovery"
#define TMPFILE "tmp"

static void save_database ()
{
	FILE *f;
	unlink (TMPFILE);
	if (!(f = fopen (TMPFILE, "w"))) {
		printf ("Can't save database\n");
		return;
	}
	savedata (f);
	fclose (f);
	unlink (DBFILE);
	link (TMPFILE, DBFILE);
	unlink (TMPFILE);
	unlink (RECOVERYFILE);	// not needed anymore
}

static bool recover ()
{
	FILE *f;
	int i = 0;
	char action [MAX_MSG], arg [MAX_MSG], sarg [MAX_MSG];

	printf ("Recovering...\n");
	if (!(f = fopen (RECOVERYFILE, "r"))) {
		printf ("Can't open recovery file ["RECOVERYFILE"]\n");
		return false;
	}
	while (1) {
		if (!fgets (action, MAX_MSG, f)) break;
		if (!fgets (arg, MAX_MSG, f)) break;
		i++;
		arg [strlen (arg) - 1] = 0;
		switch (action [0]) {
		case RACTION_DEL:
			deluser (arg);
			break;
		case RACTION_ADD:
		case RACTION_CHP:
			if (!fgets (sarg, MAX_MSG, f)) break;
			sarg [strlen (sarg) - 1] = 0;
			if (action [0] == RACTION_ADD) adduser (arg, sarg);
			else chpass (arg, sarg);
			break;
		default:
			printf ("BAD ACTION IN RECOVERY FILE."
				" ACTION %i CURRUPT!", i);
			exit (1);
		}
	}

	printf ("%i Recovery actions. Ok\n", i);

	fclose (f);

	return true;
}

static pid_t MasterPID;
void Termination (int);

short parse (int argc, char **argv)
{
	int i, j;
	char *locdir = NULL;
	short port = DEFAULT_PORT;
	char buffer [300];
	int sockfd;

	initLOCMessages ();
	umask (0177);

	for (i = 0; i < argc; i++) {
		if (argv [i][0] == '-' && argv [i][1] == 'd')
			locdir = &argv [i][2];
		else if (isdigit (argv [i][0]))
			port = strtoul (argv [i], NULL, 10);
		else {
			puts ("LOCserv" LOCSERVER_VERSION "\n"
			"Usage: LOCserv [-d<LOCDIR>] [<TCP-port>]");
			exit (1);
		}
	}

	if (!locdir) {
		if (!(locdir = getenv ("HOME"))) {
			printf ("Can't find $HOME\n");
			exit (1);
		}
		strcat (strcpy (buffer, locdir), DEFAULT_DIR);
		locdir = StrDup (buffer);
	}

	if (chdir (locdir)) {
		perror ("error:");
		printf ("Invalid directory [%s]\n", locdir);
		exit (1);
	}

	if (!access (LOCKFILE, F_OK)) {
		printf ("Lockfile [%s/"LOCKFILE"] found. Please delete this "
			"file and restart LOCserver to proceed with "
			"recovery (But make sure LOCserver is "
			"not already running!)\n", locdir);
		exit (1);
	}

	if (access (ADMINPASSFILE, R_OK)
	|| (i = open (ADMINPASSFILE, O_RDONLY)) == -1) {
		printf ("Can't read file [%s/"ADMINPASSFILE"]. This file "
			"is supposed to contain the administrator "
			"password. Please create it\n", locdir);
		exit (1);
	}

	j = read (i, buffer, MAX_PASS);
	buffer [j] = 0;
	while (buffer [strlen (buffer) - 1] == '\n')
		buffer [strlen (buffer) - 1] = 0;
	admin_password = StrDup (buffer);
	close (i);

	FILE *db = (access (DBFILE, R_OK)) ? NULL : fopen (DBFILE, "r");
	load_database (db);
	fclose (db);

	if (!access (RECOVERYFILE, R_OK))
		if (recover ()) save_database ();

	sprintf (buffer, "%i\n", MasterPID = getpid ());
//	i = open (LOCKFILE, O_WRONLY|O_CREAT);
//	write (i, buffer, strlen (buffer));
//	close (i);

	if ((sockfd = listen_tcp (port)) == -1) {
		printf ("Listen on port %i fail\n", port);
		exit (1);
	}

	server_init ();

	recovery_fd = open (RECOVERYFILE, O_WRONLY|O_CREAT, S_IRWXU);
	if (recovery_fd == -1) {
		perror ("Can't open recovery file:");
		exit (1);
	}

	signal (SIGTERM, Termination);
	signal (SIGINT, Termination);

	srand (time (NULL) ^ getpid () << 24);

	return sockfd;
}

void Termination (int)
{
	if (getpid () != MasterPID) return;
	// Bring Down the Semaphores
	sem_wait (&asem);
	sem_wait (&rsem);
	close (recovery_fd);
	save_database ();
//	unlink (LOCKFILE);
	exit (0);
}

int main (int argc, char **argv)
{
	server (parse (argc - 1, argv + 1));
}
